from typing import Dict, List, Literal

from .....utils.class_registry import register_class
from ....modifiable_values import DamageValue
from ....consts import DamageElementalType, DamageType, ObjectPositionType
from ....action import DrawCardAction, MakeDamageAction
from ....match import Match
from ....event import ReceiveDamageEventArguments, RoundEndEventArguments
from ....struct import Cost
from ....card.equipment.artifact.base import RoundEffectArtifactBase
from .....utils.desc_registry import DescDictType


class HeartOfKhvarenasBrilliance_4_3(RoundEffectArtifactBase):
    name: Literal["Heart of Khvarena's Brilliance"]
    version: Literal["4.3"] = "4.3"
    cost: Cost = Cost()
    max_usage_per_round: int = 1

    def event_handler_RECEIVE_DAMAGE(
        self, event: ReceiveDamageEventArguments, match: Match
    ) -> List[DrawCardAction]:
        if self.usage <= 0:
            # already used
            return []
        if not self.position.check_position_valid(
            event.final_damage.target_position,
            match,
            player_idx_same=True,
            character_idx_same=True,
            area_same=True,
            source_area=ObjectPositionType.CHARACTER,
            source_is_active_character=True,
        ):
            # not equipped, or not self receive damage
            return []
        character = match.player_tables[self.position.player_idx].characters[
            self.position.character_idx
        ]
        if character.hp == 0:
            # character is dying
            return []
        if event.final_damage.damage_type != DamageType.DAMAGE:
            # not damage
            return []
        # draw card
        self.usage -= 1
        return [
            DrawCardAction(
                player_idx=self.position.player_idx,
                number=1,
                draw_if_filtered_not_enough=True,
            )
        ]


class VourukashasGlow_4_3(HeartOfKhvarenasBrilliance_4_3):
    name: Literal["Vourukasha's Glow"]
    version: Literal["4.3"] = "4.3"
    cost: Cost = Cost(same_dice_number=1)

    def event_handler_ROUND_END(
        self, desc: RoundEndEventArguments, match: Match
    ) -> List[MakeDamageAction]:
        if self.position.area != ObjectPositionType.CHARACTER:
            # not equipped
            return []
        character = match.player_tables[self.position.player_idx].characters[
            self.position.character_idx
        ]
        return [
            MakeDamageAction(
                damage_value_list=[
                    DamageValue(
                        position=self.position,
                        target_position=character.position,
                        damage=-1,
                        damage_type=DamageType.HEAL,
                        damage_elemental_type=DamageElementalType.HEAL,
                        cost=Cost(),
                    )
                ]
            )
        ]


desc: Dict[str, DescDictType] = {
    "ARTIFACT/Heart of Khvarena's Brilliance": {
        "names": {"en-US": "Heart of Khvarena's Brilliance", "zh-CN": "灵光明烁之心"},
        "descs": {
            "4.3": {
                "en-US": "After this character takes DMG: Draw 1 card. (Once per Round)\n(A character can equip a maximum of 1 Artifact)",  # noqa: E501
                "zh-CN": "角色受到伤害后：如果所附属角色为「出战角色」，则抓1张牌。（每回合1次）\n（角色最多装备1件「圣遗物」）",  # noqa: E501
            }
        },
        "image_path": "cardface/Modify_Artifact_Huashen.png",  # noqa: E501
        "id": 312021,
    },
    "ARTIFACT/Vourukasha's Glow": {
        "names": {"en-US": "Vourukasha's Glow", "zh-CN": "花海甘露之光"},
        "descs": {
            "4.3": {
                "en-US": "After this character takes DMG: Draw 1 card. (Once per Round)\nEnd Phase: Heal the attached character for 1 HP.\n(A character can equip a maximum of 1 Artifact)",  # noqa: E501
                "zh-CN": "角色受到伤害后：如果所附属角色为「出战角色」，则抓1张牌。（每回合1次）\n结束阶段：治疗所附属角色1点。\n（角色最多装备1件「圣遗物」）",  # noqa: E501
            }
        },
        "image_path": "cardface/Modify_Artifact_Huashentao.png",  # noqa: E501
        "id": 312022,
    },
}


register_class(VourukashasGlow_4_3 | HeartOfKhvarenasBrilliance_4_3, desc)



# type: ignore


from typing import Dict, List, Literal

from lpsim.utils.class_registry import register_class
from lpsim.server.match import Match
from lpsim.utils.desc_registry import DescDictType
from lpsim.server.summon.base import ShieldSummonBase, AttackerSummonBase
from lpsim.server.modifiable_values import CombatActionValue, DamageIncreaseValue
from lpsim.server.event import RoundPrepareEventArguments
from lpsim.server.action import Actions, CreateObjectAction
from lpsim.server.struct import Cost, ObjectPosition
from lpsim.consts import (
    ELEMENT_TO_DAMAGE_TYPE, DamageElementalType, DamageType, DieColor, 
    ElementType, FactionType, ObjectPositionType, WeaponType
)
from lpsim.server.character.character_base import (
    ElementalBurstBase, ElementalNormalAttackBase, ElementalSkillBase, 
    PhysicalNormalAttackBase, PassiveSkillBase, CharacterBase, SkillTalent
)


# Character status.


# Round status, will last for several rounds and disappear
class ...(RoundCharacterStatus):
    name: Literal[...] = ...
    version: Literal[...] = ...
    usage: int = ...
    max_usage: int = ...

    def event_handler_...(
        self, event: ..., 
        match: Match
    ) -> List[Actions]:
        ...


# Usage status, will not disappear until usage is 0
class ...(UsageCharacterStatus):
    name: Literal[...] = ...
    version: Literal[...] = ...
    usage: int = ...
    max_usage: int = ...

    def event_handler_...(
        self, event: ..., 
        match: Match
    ) -> List[Actions]:
        ...


# Defend status, i.e. purple shield. They inherit UsageCharacterStatus
class ...(DefendCharacterStatus):
    name: Literal[...] = ...
    version: Literal[...] = ...
    usage: int = ...
    max_usage: int = ...
    min_damage_to_trigger: int = ...
    max_in_one_time: int = ...


# Shieldstatus, i.e. yellow shield.
class ...(ShieldCharacterStatus):
    name: Literal[...] = ...
    version: Literal[...] = ...
    usage: int = ...
    max_usage: int = ...


# Team status. 


# Refer to above, change XXXCharacterStatus to XXXTeamStatus.
...


# Summons


class ...(AttackerSummonBase):
    name: Literal[...] = ...
    version: Literal[...] = ...
    usage: int = ...
    max_usage: int = ...
    damage_elemental_type: DamageElementalType = DamageElementalType.PYRO
    damage: int = ...


class ...(DefendSummonBase):
    name: Literal[...] = ...
    version: Literal[...] = ...
    usage: int = ...
    max_usage: int = ...
    damage_elemental_type: DamageElementalType = DamageElementalType.PYRO
    damage: int = ...
    min_damage_to_trigger: int = ...
    max_in_one_time: int = ...
    attack_until_run_out_of_usage: bool = ...


# Skills


class ...(ElementalSkillBase):
    name: Literal[...] = ...
    damage: int = ...
    damage_type: DamageElementalType = DamageElementalType.PYRO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.PYRO,
        elemental_dice_number = 3
    )

    def get_actions(self, match: Match) -> List[Actions]:
        """
        Attack and create object
        """
        return super().get_actions(match) + [self.create_summon('...'),
            self.create_character_status('...'),
            self.create_team_status('...'),
        ]


class ...(ElementalBurstBase):
    name: Literal[...] = ...
    damage: int = ...
    damage_type: DamageElementalType = DamageElementalType.PYRO
    cost: Cost = Cost(
        elemental_dice_color = DieColor.PYRO,
        elemental_dice_number = ...
        charge = ...
    )

    def get_actions(self, match: Match) -> List[Actions]:
        ...


class ...(PassiveSkillBase):
    name: Literal[...] = ...
    usage: int = ...

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Match
    ) -> List[Actions]:
        """
        When in round prepare, reset usage
        """
        self.usage = ...
        return []

    def value_modifier_...(
        self, value: ..., 
        match: Match,
        mode: Literal['TEST', 'REAL'],
    ) -> ...:
        # triggers value modifier
        ...


# Talents


class ...(SkillTalent):
    name: Literal[...]  # Do not set default value for talent card
    version: Literal[...] = ...
    character_name: Literal[...] = ...
    cost: Cost = Cost(
        elemental_dice_color = DieColor.PYRO,
        elemental_dice_number = ...
        charge = ...
    )
    skill: Literal[...] = ...

    def event_handler_...(
        self, event: ..., 
        match: Match
    ) -> List[Actions]:
        # triggers event handler


# character base


# character class name should contain its version.
class CNAME_X_X(CharacterBase):
    name: Literal['CNAME']
    version: Literal['VERSION'] = 'VERSION'
    element: ElementType = ElementType.PYRO
    max_hp: int = 10
    max_charge: int = ...
    skills: List[
        ElementalNormalAttackBase | PhysicalNormalAttackBase
        | ...(ElementalSkillBase) | ...(ElementalBurstBase)
    ] = []
    faction: List[FactionType] = [
        FactionType.
    ]
    weapon_type: WeaponType = WeaponType.

    def _init_skills(self) -> None:
        self.skills = [
            ElementalNormalAttackBase(
                name = ...,
                damage_type = ELEMENT_TO_DAMAGE_TYPE[self.element],
                cost = ElementalNormalAttackBase.get_cost(self.element),
            ),
            PhysicalNormalAttackBase(
                name = ...,
                cost = PhysicalNormalAttackBase.get_cost(self.element),
            ),
            ...(),
            ...(),
            ...(),
        ]


# define descriptions of newly defined classes. Note key of skills and talents
# have character names. For balance changes, only descs are needed to define.
character_descs: Dict[str, DescDictType] = {
    # character information. descs are optional but must define version number.
    # id is used to sort in the frontend.
    "CHARACTER/CNAME_X_X": {
        "image_path": "character/...",
        "id": ...,
        "names": {
            "zh-CN": "...",
            "en-US": "CNAME_X_X"
        },
        "descs": {
            "VERSION": {
                "zh-CN": "",
                "en-US": ""
            }
        }
    },
    # character skills, with SKILL_${cname}_${skill_type} as key.
    "SKILL_CNAME_X_X_NORMAL_ATTACK/...": {
        "names": {
            "zh-CN": "XXX",
            "en-US": "XXX"
        },
        "descs": {
            "VERSION": {
                "zh-CN": "",
                "en-US": ""
            }
        }
    },
    ...,
    # character talents, with TALENT_${cname}_${talent_name} as key.
    "TALENT_CNAME_X_X/...": {
        "image_path": "character/...",
        "id": ...,
        "names": {
            "zh-CN": "XXX",
            "en-US": "XXX"
        },
        "descs": {
            "VERSION": {
                "zh-CN": "",
                "en-US": ""
            }
        }
    },
    # summons
    "SUMMON/...": {
        "image_path": "summon/...",
        "names": {
            "zh-CN": "XXX",
            "en-US": "XXX"
        },
        "descs": {
            "VERSION": {
                "zh-CN": "",
                "en-US": ""
            }
        }
    },
    # status
    "CHARACTER_STATUS/...": {
        "image_path": "status/...",
        "names": {
            "zh-CN": "XXX",
            "en-US": "XXX"
        },
        "descs": {
            "VERSION": {
                # you can use reference key to reference other desc
                "zh-CN": "$CHARACTER_STATUS/...|descs|OLD_VERSION",
                "en-US": ""
            }
        }
    },
}


register_class(
    CNAME_X_X | Talent | Summon | Status,
    character_descs
)



from typing import Dict, List, Literal
from lpsim.server.action import CreateDiceAction, RemoveObjectAction

from lpsim.server.card.support.locations import JadeChamber_4_0
from lpsim.server.consts import DieColor, ObjectPositionType
from lpsim.server.event import RoundPrepareEventArguments
from lpsim.server.match import Match
from lpsim.utils.class_registry import register_class
from lpsim.utils.desc_registry import DescDictType


class JadeChamber_4_5(JadeChamber_4_0):
    version: Literal["4.5"] = "4.5"

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Match
    ) -> List[CreateDiceAction | RemoveObjectAction]:
        """
        When hand number less than 3, create 1 omni die and remove self.
        """
        if (
            self.position.area != ObjectPositionType.SUPPORT
            or len(match.player_tables[self.position.player_idx].hands) > 3
        ):
            return []
        return [
            CreateDiceAction(
                player_idx=self.position.player_idx,
                number=1,
                color=DieColor.OMNI,
            ),
            RemoveObjectAction(object_position=self.position),
        ]


desc: Dict[str, DescDictType] = {
    "SUPPORT/Jade Chamber": {
        "descs": {
            "4.5": {
                "zh-CN": "投掷阶段：2个元素骰初始总是投出我方出战角色类型的元素。\n行动阶段开始时：如果我方手牌数量不多于3，则弃置此牌，生成1个万能元素。",  # noqa: E501
                "en-US": "Roll Phase: 2 initial Elemental Dice will be of the same Elemental type as your active character.\nWhen Action Phase begins: If you have no more than 3 cards in your Hand, discard this card and create 1 Omni Element.",  # noqa: E501
            }
        },
    },
}


register_class(JadeChamber_4_5, desc)



from typing import Dict, List, Literal

from .....utils.class_registry import register_class
from ....action import DrawCardAction, RemoveObjectAction
from ....match import Match
from ....event import RoundEndEventArguments
from ....consts import IconType
from ....struct import Cost
from ....card.support.locations import LocationBase
from .....utils.desc_registry import DescDictType


class WeepingWillowOfTheLake_4_3(LocationBase):
    name: Literal["Weeping Willow of the Lake"]
    version: Literal["4.3"] = "4.3"
    cost: Cost = Cost(same_dice_number=1)
    usage: int = 2
    icon_type: Literal[IconType.TIMESTATE] = IconType.TIMESTATE

    def event_handler_ROUND_END(
        self, event: RoundEndEventArguments, match: Match
    ) -> List[DrawCardAction | RemoveObjectAction]:
        """
        When in round end, if hand size less than 2, draw 2 cards, and check
        if should remove.
        """
        if self.position.area != "SUPPORT":
            # not in support area, do nothing
            return []
        if len(match.player_tables[self.position.player_idx].hands) > 2:
            # hand size more than 2, do nothing
            return []
        self.usage -= 1
        return [
            DrawCardAction(
                player_idx=self.position.player_idx,
                number=2,
                draw_if_filtered_not_enough=True,
            ),
        ] + self.check_should_remove()


desc: Dict[str, DescDictType] = {
    "SUPPORT/Weeping Willow of the Lake": {
        "names": {"en-US": "Weeping Willow of the Lake", "zh-CN": "湖中垂柳"},
        "descs": {
            "4.3": {
                "en-US": "End Phase: If your Hand has no more than 2 cards, draw 2 cards.\nUsage(s): 2",  # noqa: E501
                "zh-CN": "结束阶段：如果我方手牌数量不多于2，则抓2张牌。\n可用次数：2",  # noqa: E501
            }
        },
        "image_path": "cardface/Assist_Location_Chuiliu.png",  # noqa: E501
        "id": 321016,
    },
}


register_class(WeepingWillowOfTheLake_4_3, desc)



from lpsim.agents import InteractionAgent
from lpsim import Deck, Match, MatchState
from tests.utils_for_test import (
    check_hp,
    get_pidx_cidx,
    get_random_state,
    get_test_id_from_command,
    make_respond,
    set_16_omni,
)


def test_razor():
    cmd_records = [
        [
            "sw_card",
            "choose 0",
            "skill 1 15 14 13",
            "TEST 2 p1c2 charge 2",
            "sw_char 3 12",
            "card 2 0 11 10 9 8",
            "skill 1 7 6 5",
            "skill 2 4 3 2",
            "end",
            "skill 0 15 14 13",
            "skill 1 12 11 10",
            "sw_char 2 9",
            "card 2 0 8 7 6 5",
            "skill 2 4 3 2",
            "TEST 2 p0c1 charge 1",
            "TEST 2 p0c2 charge 0",
            "end",
            "TEST 2 p1c0 charge 1",
            "TEST 2 p1c1 charge 0",
            "sw_char 3 15",
            "skill 1 14 13 12",
        ],
        [
            "sw_card",
            "choose 2",
            "card 1 0 15 14 13 12",
            "sw_char 0 11",
            "skill 1 10 9 8",
            "TEST 2 p0c1 charge 1",
            "TEST 2 p0c3 charge 3",
            "sw_char 3 7",
            "TEST 1 7 10 10 7 4 10 7 5",
            "sw_char 0 6",
            "sw_char 3 5",
            "end",
            "choose 2",
            "sw_char 1 15",
            "TEST 1 7 10 10 7 3 4 6 0",
            "sw_char 0 14",
            "sw_char 1 13",
            "TEST 1 7 10 10 7 3 1 6 0",
            "TEST 2 p0c1 charge 1",
            "TEST 2 p0c2 charge 2",
            "end",
            "choose 2",
            "skill 1 15 14 13",
            "sw_char 0 12",
            "choose 2",
            "TEST 2 p0c1 charge 2",
            "end",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.0
        character:Chongyun
        character:Keqing
        character:Razor
        character:Razor@3.3
        Awakening*30
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0]
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                # a sample of HP check based on the command string.
                hps = cmd.strip().split(" ")[2:]
                hps = [int(x) for x in hps]
                hps = [hps[:4], hps[4:]]
                check_hp(match, hps)
            elif test_id == 2:
                cmd = cmd.split()
                pidx, cidx = get_pidx_cidx(cmd)
                assert match.player_tables[pidx].characters[cidx].charge == int(cmd[4])
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


if __name__ == "__main__":
    test_razor()



import json
from typing import Dict
from lpsim.agents import InteractionAgent
from lpsim import Deck, Match, MatchState
from tests.utils_for_test import (
    check_hp,
    check_usage,
    get_random_state,
    get_test_id_from_command,
    make_respond,
    remove_ids,
    set_16_omni,
)


def get_mamere_judgment_match():
    cmd_records = [
        [
            "sw_card",
            "choose 2",
            "card 2 0",
            "card 3 0",
            "TEST 2 p0 hand 3",
            "card 1 0",
            "TEST 2 p0 hand 4",
            "card 2 0 15 14",
            "TEST 2 p0 hand 3",
            "skill 0 13 12 11",
            "end",
            "card 3 0",
            "card 2 3 15",
            "TEST 2 p0 hand 8",
            "end",
            "card 7 0",
            "card 8 0",
            "end",
            "card 11 0",
            "card 11 2",
            "card 11 1",
            "card 6 0",
            "card 0 0 15",
            "skill 1 14 13 12",
            "end",
            "sw_char 1 15",
            "sw_char 0 14",
            "card 0 1 13 12",
            "TEST 3 p0 support 1 1 2 2",
            "TEST 2 p0 hand 16",
            "TEST 5 p0 team 3",
            "card 1 0",
            "TEST 5 p0 team 2",
            "card 1 0 11",
            "TEST 5 p0 team 1",
            "card 2 0 10",
            "TEST 5 p0 team",
            "TEST 4 p0 dice 10",
            "card 0 0",
            "TEST 1 10 8 7 9 8 10",
            "end",
        ],
        [
            "sw_card",
            "choose 0",
            "card 4 0",
            "card 1 0 15",
            "TEST 2 p1 hand 5",
            "card 2 0",
            "TEST 2 p1 hand 5",
            "card 3 0",
            "card 1 0",
            "TEST 2 p1 hand 4",
            "end",
            "card 3 0 15",
            "end",
            "card 1 0",
            "card 8 0",
            "card 7 0",
            "end",
            "card 12 0",
            "card 12 0",
            "card 12 1",
            "sw_char 1 15",
            "card 5 1",
            "card 4 1",
            "TEST 1 10 10 10 9 8 10",
            "end",
            "card 0 0 15",
            "skill 1 14 13 12",
            "skill 1 11 10 9",
            "card 2 0 8",
            "TEST 2 p1 hand 17",
            "sw_char 0 7",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.3
        character:Baizhu
        character:Nilou
        character:Tighnari
        Mamere*10
        Strategize*3
        Sweet Madame*3
        Liben*3
        Wangshu Inn*3
        Passing of Judgment*1
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    match.config.max_hand_size = 999
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    return agent_0, agent_1, match


def test_mamere_judgment():
    """
    Mamare: 3 usage, Judgment: 3 usage
    Judgment can judge arcane, food, and event card.
    """
    agent_0, agent_1, match = get_mamere_judgment_match()
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0].strip().split(" ")
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                # a sample of HP check based on the command string.
                hps = cmd[2:]
                hps = [int(x) for x in hps]
                hps = [hps[:3], hps[3:]]
                check_hp(match, hps)
            elif test_id == 2:
                pidx = int(cmd[2][1])
                assert int(cmd[-1]) == len(match.player_tables[pidx].hands)
            elif test_id == 3:
                pidx = int(cmd[2][1])
                check_usage(match.player_tables[pidx].supports, cmd[4:])
            elif test_id == 4:
                assert len(match.player_tables[0].dice.colors) == 10
            elif test_id == 5:
                pidx = int(cmd[2][1])
                check_usage(match.player_tables[pidx].team_status, cmd[4:])
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_judgment_one_round():
    cmd_records = [
        ["sw_card", "choose 0", "card 0 0 0", "end", "TEST 3 p1 usage", "end"],
        [
            "sw_card",
            "choose 1",
            "TEST 3 p1 usage 3",
            "end",
            "card 1 0 15",
            "TEST 2 p1 hand 8",
            "end",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.3
        character:Baizhu
        character:Nilou
        character:Tighnari
        Mamere*10
        Strategize*3
        Sweet Madame*3
        Liben*3
        Wangshu Inn*3
        Passing of Judgment*1
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    match.config.max_hand_size = 999
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0].strip().split(" ")
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 2:
                pidx = int(cmd[2][1])
                assert int(cmd[-1]) == len(match.player_tables[pidx].hands)
            elif test_id == 3:
                pidx = int(cmd[2][1])
                check_usage(match.player_tables[pidx].team_status, cmd[4:])
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_judge_arcane():
    cmd_records = [
        ["sw_card", "choose 0", "card 0 0 0", "end", "end", "TEST 3 p1 usage", "end"],
        [
            "sw_card",
            "choose 1",
            "card 0 0 15",
            "TEST 3 p1 usage 2",
            "TEST 3 p0 usage",
            "end",
            "end",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.3
        character:Baizhu
        character:Nilou
        character:Tighnari
        Mamere*10
        Strategize*3
        Sweet Madame*3
        Liben*3
        Wangshu Inn*3
        Passing of Judgment*1
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    match.config.max_hand_size = 999
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0].strip().split(" ")
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 3:
                pidx = int(cmd[2][1])
                check_usage(match.player_tables[pidx].team_status, cmd[4:])
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_mamere_no_default():
    """
    When mamere generate card, then success. No need to check or use the card,
    as we don't know the newest version of cards.
    """
    cmd_records = [
        [
            "sw_card",
            "choose 2",
            "card 2 0",
            "card 3 0",
            "TEST 2 p0 hand 3",
            "card 1 0",
            "TEST 2 p0 hand 4",
            "end",
        ],
        [
            "sw_card",
            "choose 0",
            "end",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        character:Baizhu
        character:Nilou
        character:Tighnari
        Mamere*10
        Strategize*3
        Sweet Madame*3
        Liben*3
        Wangshu Inn*3
        Passing of Judgment*1
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    match.config.max_hand_size = 999
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0].strip().split(" ")
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 2:
                pidx = int(cmd[2][1])
                assert int(cmd[-1]) == len(match.player_tables[pidx].hands)
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def test_mamere_judgment_2():
    """
    When food card is disabled by judgment, it still can call mamere to draw
    card.
    """
    cmd_records = [
        [
            "sw_card",
            "choose 1",
            "card 0 0 15",
            "card 1 0",
            "card 2 0",
            "skill 1 14 13 12",
            "card 0 0",
            "TEST 1 10 8 10 10 7 10",
            "TEST 2 p0 hand 3",
            "TEST 2 p1 hand 0",
            "end",
        ],
        [
            "sw_card",
            "choose 1",
            "card 0 0 15",
            "card 1 0",
            "card 2 0",
            "card 0 0",
            "card 0 0",
            "card 0 0",
            "card 0 0 14",
            "skill 1 13 12 11",
        ],
    ]
    agent_0 = InteractionAgent(
        player_idx=0, verbose_level=0, commands=cmd_records[0], only_use_command=True
    )
    agent_1 = InteractionAgent(
        player_idx=1, verbose_level=0, commands=cmd_records[1], only_use_command=True
    )
    # initialize match. It is recommended to use default random state to make
    # replay unchanged.
    match = Match(random_state=get_random_state())
    # deck information
    deck = Deck.from_str(
        """
        default_version:4.3
        character:Kaedehara Kazuha
        character:Klee
        character:Kaeya
        Passing of Judgment
        Mamere*10
        Sweet Madame*10
        """
    )
    match.set_deck([deck, deck])
    match.config.max_same_card_number = None
    match.config.character_number = None
    match.config.card_number = None
    match.config.check_deck_restriction = False
    match.config.max_hand_size = 999
    # check whether random_first_player is enabled.
    match.config.random_first_player = False
    # check whether in rich mode (16 omni each round)
    set_16_omni(match)
    match.start()
    match.step()

    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        # do tests
        while True:
            cmd = agent.commands[0].strip().split(" ")
            test_id = get_test_id_from_command(agent)
            if test_id == 0:
                # id 0 means current command is not a test command.
                break
            elif test_id == 1:
                # a sample of HP check based on the command string.
                hps = cmd[2:]
                hps = [int(x) for x in hps]
                hps = [hps[:3], hps[3:]]
                check_hp(match, hps)
            elif test_id == 2:
                pidx = int(cmd[2][1])
                assert int(cmd[-1]) == len(match.player_tables[pidx].hands)
            else:
                raise AssertionError(f"Unknown test id {test_id}")
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break

    # simulate ends, check final state
    assert match.state != MatchState.ERROR


def remove_ellin_information(match_dict: Dict):
    # remove Ellin information
    h = match_dict
    eh = h["event_handlers"]
    if len(eh) < 3:
        return json.dumps(match_dict)
    ellin_handler = h["event_handlers"][2]
    assert ellin_handler["name"] == "Ellin"
    ellin_handler["recorded_skill_ids"] = {}
    return json.dumps(h)


def test_mamere_load_and_save():
    agent_0, agent_1, match = get_mamere_judgment_match()
    assert match.start()[0]
    match.step()
    histories = [remove_ellin_information(match.dict())]
    while True:
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        while agent.commands[0].startswith("TEST"):
            agent.commands = agent.commands[1:]
        # respond
        make_respond(agent, match)
        histories.append(remove_ellin_information(match.dict()))
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break
    agent_0, agent_1, match = get_mamere_judgment_match()
    assert match.start()[0]
    match.step()
    for idx, old_match_json in enumerate(histories):  # pragma: no branch
        match_copy = match.copy(deep=True)
        assert remove_ids(match_copy) == remove_ids(
            Match(**json.loads(match_copy.json()))
        )
        assert remove_ids(
            Match(**json.loads(remove_ellin_information(match_copy.dict())))
        ) == remove_ids(Match(**json.loads(old_match_json)))
        match = Match(**json.loads(match.json()))
        if match.need_respond(0):
            agent = agent_0
        elif match.need_respond(1):
            agent = agent_1
        else:
            raise AssertionError("No need respond.")
        while agent.commands[0].startswith("TEST"):
            agent.commands = agent.commands[1:]
        # respond
        make_respond(agent, match)
        if len(agent_1.commands) == 0 and len(agent_0.commands) == 0:
            break
    assert match.state != MatchState.ERROR


if __name__ == "__main__":
    # test_mamere_judgment()
    # test_judgment_one_round()
    # test_judge_arcane()
    # test_mamere_no_default()
    # test_mamere_judgment_2()
    test_mamere_load_and_save()



from typing import Dict, List, Literal

from .....utils.class_registry import register_class
from ....consts import ELEMENT_TO_DIE_COLOR, ObjectPositionType, SkillType
from ....character.character_base import SkillBase
from ....action import Actions, CreateDiceAction, DrawCardAction
from ....match import Match
from ....event import RoundPrepareEventArguments, SkillEndEventArguments
from ....struct import Cost
from ....card.equipment.artifact.base import RoundEffectArtifactBase
from .....utils.desc_registry import DescDictType


class FlowingRings_4_3(RoundEffectArtifactBase):
    name: Literal["Flowing Rings"]
    version: Literal["4.3"] = "4.3"
    cost: Cost = Cost()
    max_usage_per_round: int = 1

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Match
    ) -> List[DrawCardAction]:
        if self.usage == 0:
            # already used
            return []
        if not self.position.check_position_valid(
            event.action.position,
            match,
            player_idx_same=True,
            character_idx_same=True,
            target_area=ObjectPositionType.SKILL,
            source_area=ObjectPositionType.CHARACTER,
        ):
            # not self, or not equipped
            return []
        skill: SkillBase = match.get_object(event.action.position)  # type: ignore
        assert skill is not None
        if skill.skill_type != SkillType.NORMAL_ATTACK:
            # not normal attack
            return []
        self.usage -= 1
        return [
            DrawCardAction(
                player_idx=self.position.player_idx,
                number=1,
                draw_if_filtered_not_enough=True,
            )
        ]


class EchoesOfAnOffering_4_3(FlowingRings_4_3):
    name: Literal["Echoes of an Offering"]
    version: Literal["4.3"] = "4.3"
    cost: Cost = Cost(same_dice_number=2)
    max_usage_per_round: int = 1

    generate_die_usage: int = 1
    max_generate_die_usage_per_round: int = 1

    def equip(self, match: Match) -> List[Actions]:
        self.usage = self.max_usage_per_round
        self.generate_die_usage = self.max_generate_die_usage_per_round
        return []

    def event_handler_ROUND_PREPARE(
        self, event: RoundPrepareEventArguments, match: Match
    ) -> List[Actions]:
        self.usage = self.max_usage_per_round
        self.generate_die_usage = self.max_generate_die_usage_per_round
        return []

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Match
    ) -> List[DrawCardAction | CreateDiceAction]:
        ret: List[DrawCardAction | CreateDiceAction] = []
        ret += super().event_handler_SKILL_END(event, match)
        if self.generate_die_usage == 0:
            # already used
            return ret
        if not self.position.check_position_valid(
            event.action.position,
            match,
            player_idx_same=True,
            character_idx_same=True,
            target_area=ObjectPositionType.SKILL,
            source_area=ObjectPositionType.CHARACTER,
        ):
            # not self, or not equipped
            return ret
        table = match.player_tables[self.position.player_idx]
        if len(table.dice.colors) > len(table.hands):
            # dice more than hand
            return ret
        character = table.characters[self.position.character_idx]
        # create die
        self.generate_die_usage -= 1
        ret.append(
            CreateDiceAction(
                player_idx=self.position.player_idx,
                color=ELEMENT_TO_DIE_COLOR[character.element],
                number=1,
            )
        )
        return ret


desc: Dict[str, DescDictType] = {
    "ARTIFACT/Flowing Rings": {
        "names": {"en-US": "Flowing Rings", "zh-CN": "浮溯之珏"},
        "descs": {
            "4.3": {
                "en-US": "After the character uses a Normal Attack: Draw 1 card. (Once per Round)\n(A character can equip a maximum of 1 Artifact)",  # noqa: E501
                "zh-CN": "角色使用「普通攻击」后：抓1张牌。（每回合1次）\n（角色最多装备1件「圣遗物」）",  # noqa: E501
            }
        },
        "image_path": "cardface/Modify_Artifact_Yuxiang.png",  # noqa: E501
        "id": 312019,
    },
    "ARTIFACT/Echoes of an Offering": {
        "names": {"en-US": "Echoes of an Offering", "zh-CN": "来歆余响"},
        "descs": {
            "4.3": {
                "en-US": "After this character uses Normal Attack: Draw 1 card. (Once per Round)\nAfter this character uses a Skill: If you do not have more Elemental Dice than cards in your Hand, you will gain 1 Elemental Die of the attached character's Type. (Once per Round)\n(A character can equip a maximum of 1 Artifact)",  # noqa: E501
                "zh-CN": "角色使用「普通攻击」后：抓1张牌。（每回合1次）\n角色使用技能后：如果我方元素骰数量不多于手牌数量，则生成1个所附属角色类型的元素骰。（每回合1次）\n（角色最多装备1件「圣遗物」）",  # noqa: E501
            }
        },
        "image_path": "cardface/Modify_Artifact_Yuxiangtao.png",  # noqa: E501
        "id": 312020,
    },
}


register_class(FlowingRings_4_3 | EchoesOfAnOffering_4_3, desc)



from ...utils import import_all_modules
from .equipment import WeaponBase, ArtifactBase
from .support import SupportBase, LocationBase, ItemBase, CompanionBase


import_all_modules(__file__, __name__)
__all__ = (
    "WeaponBase",
    "ArtifactBase",
    "SupportBase",
    "LocationBase",
    "ItemBase",
    "CompanionBase",
)



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...struct import Cost

from ...action import (
    Actions,
    ChangeObjectUsageAction,
    MakeDamageAction,
    RemoveObjectAction,
    SkillEndAction,
)

from ...event import (
    ChangeObjectUsageEventArguments,
    MakeDamageEventArguments,
    RoundEndEventArguments,
    SkillEndEventArguments,
)

from ...consts import (
    DamageElementalType,
    DamageType,
    DieColor,
    ElementType,
    IconType,
    ObjectPositionType,
    SkillType,
)

from ...modifiable_values import (
    CostValue,
    DamageElementEnhanceValue,
    DamageIncreaseValue,
    DamageValue,
)
from .base import (
    ElementalInfusionCharacterStatus,
    PrepareCharacterStatus,
    ReviveCharacterStatus,
    RoundCharacterStatus,
    ShieldCharacterStatus,
    UsageCharacterStatus,
    CharacterStatusBase,
)


class ElectroInfusionKeqing_3_3(ElementalInfusionCharacterStatus, RoundCharacterStatus):
    """
    Inherit from vanilla elemental infusion status, and has talent activated
    argument. when talent activated, it will gain electro damage +1 buff,
    and usage +1.
    """

    name: Literal["Electro Elemental Infusion"] = "Electro Elemental Infusion"
    desc: Literal["", "keqing_talent"] = ""
    mark: Literal["Keqing"]  # used to select right status
    version: Literal["3.3"] = "3.3"
    usage: int = 2
    max_usage: int = 2

    talent_activated: bool = False

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        if self.talent_activated:
            self.desc = "keqing_talent"

    def renew(self, new_status: "ElectroInfusionKeqing_3_3") -> None:
        super().renew(new_status)
        if new_status.talent_activated:
            self.talent_activated = True
            self.desc = "keqing_talent"

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any, mode: Literal["TEST", "REAL"]
    ) -> DamageIncreaseValue:
        """
        If talent activated, increase electro damage dealt by this character
        by 1.
        """
        if not self.talent_activated:
            # talent not activated, do nothing
            return value
        if not value.is_corresponding_character_use_damage_skill(
            self.position, match, None
        ):
            # not corresponding character use skill, do nothing
            return value
        if value.damage_elemental_type != DamageElementalType.ELECTRO:
            # not electro damage, do nothing
            return value
        value.damage += 1
        return value


class ElectroCrystalCore_3_7(ReviveCharacterStatus):
    name: Literal["Electro Crystal Core"] = "Electro Crystal Core"
    version: Literal["3.7"] = "3.7"
    heal: int = 1


class RockPaperScissorsComboScissors_3_7(PrepareCharacterStatus):
    name: Literal[
        "Rock-Paper-Scissors Combo: Scissors"
    ] = "Rock-Paper-Scissors Combo: Scissors"
    version: Literal["3.7"] = "3.7"
    character_name: Literal["Electro Hypostasis"] = "Electro Hypostasis"
    skill_name: Literal[
        "Rock-Paper-Scissors Combo: Scissors"
    ] = "Rock-Paper-Scissors Combo: Scissors"


class RockPaperScissorsComboPaper_3_7(PrepareCharacterStatus):
    name: Literal[
        "Rock-Paper-Scissors Combo: Paper"
    ] = "Rock-Paper-Scissors Combo: Paper"
    version: Literal["3.7"] = "3.7"
    character_name: Literal["Electro Hypostasis"] = "Electro Hypostasis"
    skill_name: Literal[
        "Rock-Paper-Scissors Combo: Paper"
    ] = "Rock-Paper-Scissors Combo: Paper"


class ChakraDesiderata_3_7(CharacterStatusBase):
    name: Literal["Chakra Desiderata"] = "Chakra Desiderata"
    version: Literal["3.7"] = "3.7"
    usage: int = 0
    max_usage: int = 3
    icon_type: Literal[IconType.OTHERS] = IconType.OTHERS

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[Actions]:
        """
        When other character using elemental burst, gain resolve.
        """
        if not self.position.check_position_valid(
            event.action.position,
            match,
            player_idx_same=True,
            character_idx_same=False,
            target_area=ObjectPositionType.SKILL,
        ):
            # not other ally use elemental burst, do nothing
            return []
        if event.action.skill_type == SkillType.ELEMENTAL_BURST:
            # is burst
            self.usage = min(self.usage + 1, self.max_usage)
        return []

    def value_modifier_DAMAGE_INCREASE(
        self, value: DamageIncreaseValue, match: Any, mode: Literal["TEST", "REAL"]
    ) -> DamageIncreaseValue:
        """
        When self using elemental burst, increase damage by usage.
        """
        if not value.is_corresponding_character_use_damage_skill(
            self.position, match, SkillType.ELEMENTAL_BURST
        ):
            # not self use burst, do nothing
            return value
        # increase damage
        value.damage += self.usage
        character = match.player_tables[self.position.player_idx].characters[
            self.position.character_idx
        ]
        if character.talent is not None:
            # has talent, double increase damage
            value.damage += self.usage
        self.usage = 0
        return value


class TheShrinesSacredShade_3_7(RoundCharacterStatus):
    name: Literal["The Shrine's Sacred Shade"] = "The Shrine's Sacred Shade"
    version: Literal["3.7"] = "3.7"
    usage: int = 1
    max_usage: int = 1
    icon_type: Literal[IconType.SPECIAL] = IconType.SPECIAL

    def value_modifier_COST(
        self,
        value: CostValue,
        match: Any,
        mode: Literal["TEST", "REAL"],
    ) -> CostValue:
        """
        When use Yakan Evocation: Sesshou Sakura, cost 2 less
        """
        if not self.position.check_position_valid(
            value.position,
            match,
            player_idx_same=True,
            character_idx_same=True,
            target_area=ObjectPositionType.SKILL,
        ):
            # not our player skill, do nothing
            return value
        skill = match.get_object(value.position)
        if skill.name != "Yakan Evocation: Sesshou Sakura":
            # not right skill, do nothing
            return value
        assert self.usage > 0
        success = value.cost.decrease_cost(DieColor.ELECTRO)
        success = value.cost.decrease_cost(DieColor.ELECTRO) or success
        if success and mode == "REAL":
            self.usage -= 1
        return value

    def event_handler_SKILL_END(
        self, event: SkillEndAction, match: Any
    ) -> List[RemoveObjectAction]:
        return self.check_should_remove()


class TheWolfWithin_3_3(RoundCharacterStatus):
    name: Literal["The Wolf Within"] = "The Wolf Within"
    version: Literal["3.3"] = "3.3"
    usage: int = 2
    max_usage: int = 2
    icon_type: Literal[IconType.OTHERS] = IconType.OTHERS

    def event_handler_SKILL_END(
        self, event: SkillEndEventArguments, match: Any
    ) -> List[MakeDamageAction]:
        """
        If self use normal attack or elemental skill, deal damage
        """
        if not self.position.check_position_valid(
            event.action.position,
            match,
            player_idx_same=True,
            character_idx_same=True,
            target_area=ObjectPositionType.SKILL,
        ):
            # not self use skill
            return []
        if event.action.skill_type not in [
            SkillType.NORMAL_ATTACK,
            SkillType.ELEMENTAL_SKILL,
        ]:
            # not normal attack or elemental skill
            return []
        # make damage
        target = match.player_tables[
            1 - self.position.player_idx
        ].get_active_character()
        return [
            MakeDamageAction(
                damage_value_list=[
                    DamageValue(
                        position=self.position,
                        damage_type=DamageType.DAMAGE,
                        target_position=target.position,
                        damage=2,
                        damage_elemental_type=DamageElementalType.ELECTRO,
                        cost=Cost(),
                    )
                ]
            )
        ]


class TidecallerSurfEmbrace_3_4(ShieldCharacterStatus, PrepareCharacterStatus):
    name: Literal["Tidecaller: Surf Embrace"] = "Tidecaller: Surf Embrace"
    version: Literal["3.4"] = "3.4"
    character_name: Literal["Beidou"] = "Beidou"
    skill_name: Literal["Wavestrider"] = "Wavestrider"

    usage: int = 2
    max_usage: int = 2

    def event_handler_MAKE_DAMAGE(
        self, event: MakeDamageEventArguments, match: Any
    ) -> List[Actions]:
        """
        Do not remove when usage becomes zero
        """
        return []


class CrowfeatherCover_3_5(UsageCharacterStatus):
    name: Literal["Crowfeather Cover"] = "Crowfeather Cover"
    version: Literal["3.5"] = "3.5"
    usage: int = 2
    max_usage: int = 2
    icon_type: Literal[IconType.ATK_UP] = IconType.ATK_UP

    def value_modifier_DAMAGE_INCREASE(
        self,
        value: DamageIncreaseValue,
        match: Any,
        mode: Literal["TEST", "REAL"],
    ) -> DamageIncreaseValue:
        if value.damage_from_element_reaction:
            # is elemental reaction, do nothing
            return value
        if value.is_corresponding_character_use_damage_skill(
            self.position, match, SkillType.ELEMENTAL_SKILL
        ) or value.is_corresponding_character_use_damage_skill(
            self.position, match, SkillType.ELEMENTAL_BURST
        ):
            # is self use elemental skill or elemental burst
            if self.usage <= 0:  # pragma: no cover
                # no usage
                return value
            # increase damage
            assert mode == "REAL"
            self.usage -= 1
            value.damage += 1
            # talent effects
            characters = match.player_tables[self.position.player_idx].characters
            if characters[self.position.character_idx].element != ElementType.ELECTRO:
                # not electro
                return value
            found_talent_kujou: bool = False
            for character in characters:
                if character.name == "Kujou Sara" and character.talent is not None:
                    found_talent_kujou = True
                    break
            if found_talent_kujou:
                # increase one more damage
                value.damage += 1
        return value


class PactswornPathclearer_3_3(ElementalInfusionCharacterStatus):
    name: Literal["Pactsworn Pathclearer"] = "Pactsworn Pathclearer"
    version: Literal["3.3"] = "3.3"
    usage: int = 0
    max_usage: int = 999
    infused_elemental_type: DamageElementalType = DamageElementalType.ELECTRO
    icon_type: Literal[IconType.OTHERS] = IconType.OTHERS

    def value_modifier_DAMAGE_ELEMENT_ENHANCE(
        self,
        value: DamageElementEnhanceValue,
        match: Any,
        mode: Literal["TEST", "REAL"],
    ) -> DamageElementEnhanceValue:
        if self.usage < 2:
            # level not enough
            return value
        return super().value_modifier_DAMAGE_ELEMENT_ENHANCE(value, match, mode)

    def value_modifier_DAMAGE_INCREASE(
        self,
        value: DamageIncreaseValue,
        match: Any,
        mode: Literal["TEST", "REAL"],
    ) -> DamageIncreaseValue:
        if self.usage < 4:
            # level not enough
            return value
        if value.is_corresponding_character_use_damage_skill(
            self.position, match, None
        ):
            # self use skill, increase damage
            assert mode == "REAL"
            value.damage += 2
        return value

    def event_handler_ROUND_END(
        self, event: RoundEndEventArguments, match: Any
    ) -> List[ChangeObjectUsageAction]:
        """
        add one usage
        """
        return [ChangeObjectUsageAction(object_position=self.position, change_usage=1)]

    def event_handler_CHANGE_OBJECT_USAGE(
        self, event: ChangeObjectUsageEventArguments, match: Any
    ) -> List[Actions]:
        """
        When change object usage, check if usage is greater than 6.
        If so, decrease by 4.
        """
        if self.usage >= 6:
            self.usage -= 4
        return []


class Conductive_4_0(CharacterStatusBase):
    """
    repeatedly attach, end phase accumulate will perform in this status;
    damage increase will perform in elemental skill.
    """

    name: Literal["Conductive"] = "Conductive"
    version: Literal["4.0"] = "4.0"
    usage: int = 2
    max_usage: int = 4
    icon_type: Literal[IconType.DEBUFF] = IconType.DEBUFF

    def renew(self, new_status: "Conductive_4_0"):
        """
        Add one usage
        """
        self.usage = min(self.max_usage, self.usage + 1)

    def event_handler_ROUND_END(
        self, event: RoundEndEventArguments, match: Any
    ) -> List[Actions]:
        """
        When in round end, add one usage
        """
        self.usage = min(self.max_usage, self.usage + 1)
        return []


register_class(
    ElectroInfusionKeqing_3_3
    | RockPaperScissorsComboScissors_3_7
    | RockPaperScissorsComboPaper_3_7
    | ElectroCrystalCore_3_7
    | ChakraDesiderata_3_7
    | TheShrinesSacredShade_3_7
    | TheWolfWithin_3_3
    | TidecallerSurfEmbrace_3_4
    | CrowfeatherCover_3_5
    | PactswornPathclearer_3_3
    | Conductive_4_0
)



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...action import Actions
from ...struct import Cost

from ...consts import (
    DamageElementalType,
    DieColor,
    ElementType,
    FactionType,
    WeaponType,
)
from ..character_base import (
    ElementalBurstBase,
    ElementalSkillBase,
    PhysicalNormalAttackBase,
    CharacterBase,
    SkillTalent,
)


# Character status. DO NOT define here, define in server/status/characor_status
# Here is just example.


# Round status, will last for several rounds and disappear
# Skills


class ChonghuasLayeredFrost(ElementalSkillBase):
    name: Literal["Chonghua's Layered Frost"] = "Chonghua's Layered Frost"
    damage: int = 3
    damage_type: DamageElementalType = DamageElementalType.CRYO
    cost: Cost = Cost(elemental_dice_color=DieColor.CRYO, elemental_dice_number=3)

    def get_actions(self, match: Any) -> List[Actions]:
        """
        Attack and create object
        """
        args = {}
        character = match.player_tables[self.position.player_idx].characters[
            self.position.character_idx
        ]
        talent = character.talent
        if talent is not None:
            args = {
                "talent_activated": True,
                "usage": talent.status_max_usage,
                "max_usage": talent.status_max_usage,
            }
        return super().get_actions(match) + [
            self.create_team_status("Chonghua's Frost Field", args),
        ]


# Talents


class SteadyBreathing_3_3(SkillTalent):
    name: Literal["Steady Breathing"]
    version: Literal["3.3"] = "3.3"
    character_name: Literal["Chongyun"] = "Chongyun"
    cost: Cost = Cost(elemental_dice_color=DieColor.CRYO, elemental_dice_number=4)
    skill: Literal["Chonghua's Layered Frost"] = "Chonghua's Layered Frost"
    status_max_usage: int = 3


class SteadyBreathing_4_2(SkillTalent):
    name: Literal["Steady Breathing"]
    version: Literal["4.2"] = "4.2"
    character_name: Literal["Chongyun"] = "Chongyun"
    cost: Cost = Cost(elemental_dice_color=DieColor.CRYO, elemental_dice_number=3)
    skill: Literal["Chonghua's Layered Frost"] = "Chonghua's Layered Frost"
    status_max_usage: int = 2


# character base


class Chongyun_3_3(CharacterBase):
    name: Literal["Chongyun"]  # Do not set default value for character name
    version: Literal["3.3"] = "3.3"
    element: ElementType = ElementType.CRYO
    max_hp: int = 10
    max_charge: int = 3
    skills: List[
        PhysicalNormalAttackBase | ChonghuasLayeredFrost | ElementalBurstBase
    ] = []
    faction: List[FactionType] = [FactionType.LIYUE]
    weapon_type: WeaponType = WeaponType.CLAYMORE

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name="Demonbane",
                cost=PhysicalNormalAttackBase.get_cost(self.element),
            ),
            ChonghuasLayeredFrost(),
            ElementalBurstBase(
                name="Cloud-Parting Star",
                damage=7,
                damage_type=DamageElementalType.CRYO,
                cost=ElementalBurstBase.get_cost(self.element, 3, 3),
            ),
        ]


register_class(Chongyun_3_3 | SteadyBreathing_3_3 | SteadyBreathing_4_2)



from typing import Any, List, Literal

from ....utils.class_registry import register_class

from ...event import (
    AfterMakeDamageEventArguments,
    CharacterDefeatedEventArguments,
)

from ...action import CreateObjectAction
from ...struct import Cost

from ...consts import (
    DamageElementalType,
    DieColor,
    ElementType,
    FactionType,
    ObjectPositionType,
    WeaponType,
)
from ..character_base import (
    CreateStatusPassiveSkill,
    ElementalBurstBase,
    ElementalSkillBase,
    PhysicalNormalAttackBase,
    CharacterBase,
    SkillTalent,
)


# Skills


class UpaShato(ElementalBurstBase):
    name: Literal["Upa Shato"] = "Upa Shato"
    damage: int = 5
    damage_type: DamageElementalType = DamageElementalType.PHYSICAL
    cost: Cost = Cost(
        elemental_dice_color=DieColor.GEO, elemental_dice_number=3, charge=2
    )


class InfusedStonehide(CreateStatusPassiveSkill):
    name: Literal["Infused Stonehide"] = "Infused Stonehide"
    status_name: Literal["Stonehide"] = "Stonehide"


# Talents


class StonehideReforged_3_3(SkillTalent):
    name: Literal["Stonehide Reforged"]
    version: Literal["3.3"] = "3.3"
    character_name: Literal["Stonehide Lawachurl"] = "Stonehide Lawachurl"
    cost: Cost = Cost(
        elemental_dice_color=DieColor.GEO, elemental_dice_number=4, charge=2
    )
    skill: Literal["Upa Shato"] = "Upa Shato"

    opposite_alive: List[int] = []

    def event_handler_AFTER_MAKE_DAMAGE(
        self, event: AfterMakeDamageEventArguments, match: Any
    ):
        """
        record enemy alive status
        """
        if self.position.area != ObjectPositionType.CHARACTER:
            # not equipped, return
            return []
        # reset alive
        self.opposite_alive = []
        assert len(event.action.damage_value_list) > 0
        # if use skill, source of first should be self
        if not self.position.check_position_valid(
            event.action.damage_value_list[0].position,
            match,
            player_idx_same=True,
            character_idx_same=True,
            target_area=ObjectPositionType.SKILL,
        ):
            # not self use skill, return
            return []
        for cid, character in enumerate(
            match.player_tables[1 - self.position.player_idx].characters
        ):
            self.opposite_alive.append(cid)
        return []

    def event_handler_CHARACTER_DEFEATED(
        self, event: CharacterDefeatedEventArguments, match: Any
    ) -> List[CreateObjectAction]:
        """
        If self is using skill, and enemy defeated, re-attach Stonehide and
        Stone Force.
        """
        if self.position.area != ObjectPositionType.CHARACTER:
            # not equipped, return
            return []
        if event.action.player_idx == self.position.player_idx:
            # self defeated, return
            return []
        if event.action.character_idx not in self.opposite_alive:
            # defeated character not in opposite_alive, return
            return []
        # re-attach Stonehide and Stone Force
        return [
            CreateObjectAction(
                object_name="Stonehide",
                object_position=self.position.set_area(
                    ObjectPositionType.CHARACTER_STATUS
                ),
                object_arguments={},
            )
        ]


# character base


class StonehideLawachurl_3_3(CharacterBase):
    name: Literal["Stonehide Lawachurl"]
    version: Literal["3.3"] = "3.3"
    element: ElementType = ElementType.GEO
    max_hp: int = 8
    max_charge: int = 2
    skills: List[
        PhysicalNormalAttackBase | ElementalSkillBase | UpaShato | InfusedStonehide
    ] = []
    faction: List[FactionType] = [FactionType.MONSTER, FactionType.HILICHURL]
    weapon_type: WeaponType = WeaponType.OTHER

    def _init_skills(self) -> None:
        self.skills = [
            PhysicalNormalAttackBase(
                name="Plama Lawa",
                cost=PhysicalNormalAttackBase.get_cost(self.element),
            ),
            ElementalSkillBase(
                name="Movo Lawa",
                damage_type=DamageElementalType.PHYSICAL,
                cost=ElementalSkillBase.get_cost(self.element),
            ),
            UpaShato(),
            InfusedStonehide(),
        ]


register_class(StonehideLawachurl_3_3 | StonehideReforged_3_3)



"""
parse instance based on base class. When an object is registered, we will scan
all its parent class, and register it to the corresponding base class. When we
need to instantiate an object with a dict, with given base class, we will find
corresponding class and version, and instantiate it.
"""


import types
from typing import Any, Dict, List, Optional, Set, Union

from .instance_factory import InstanceFactory
from .desc_registry import DescDictType, update_desc

instance_factory = InstanceFactory()


def _is_union_type(t) -> bool:
    return (
        getattr(t, "__class__", None) == types.UnionType
        or getattr(t, "__origin__", None) is Union
        or getattr(t, "__origin__", None) is Optional
    )


def register_base_class(base_class: Any):
    """
    Register a base class.
    TODO: after refactor, we should remove this function. Currently remain here for
    compatibility.
    """
    return


def register_class(classes: Any, descs: Dict[str, DescDictType] | None = None):
    """
    Register classes with their descriptions. If classes is a Union, register
    all classes in the union. Otherwise, register the class itself.
    Before registering, descs will be updated first.
    """
    if descs is not None:
        update_desc(descs)
    if _is_union_type(classes):
        for class_type in classes.__args__:  # type: ignore
            instance_factory.register_instance(class_type)
    else:
        instance_factory.register_instance(classes)


def get_instance(base_class: Any, args: Dict):
    """
    Get instance from registry. If the base class is a union type, we will try
    each type sequentially.
    """
    base_class_list = [base_class]
    if _is_union_type(base_class):
        base_class_list = base_class.__args__  # type: ignore
    for cls in base_class_list:
        try:
            return instance_factory.get_instance(cls, args)
        except KeyError:
            continue

    names = [x.__name__ for x in base_class_list]
    raise AssertionError(
        "Instance %s not found in instance factory with args %s" % (names, args)
    )


def get_class_list_by_base_class(
    base_class: Any, version: str = "99.9", exclude: Set[str] = set()
) -> List[str]:
    """
    Get list of class names by base class and version. If the base class is a
    union type, class names that matches any type in the union will be returned.

    Args:
        base_class: The base class to search.
        version: The version of the class. Classes that have higher version
            than this will be ignored. Default is '99.9'.
        exclude: Exclude class with specified names in the list.

    Return:
        List of class names.
    """
    result_set: Set[str] = set()
    base_class_list = [base_class]
    if _is_union_type(base_class):
        # is union type, try each class sequentially
        base_class_list = base_class.__args__  # type: ignore
    for type in base_class_list:
        for key in instance_factory.instance_register.keys():
            card_name = key.split("+")[1]
            card_version = key.split("+")[2]
            if card_name in exclude:
                continue
            if issubclass(instance_factory.instance_register[key], type):
                if card_version <= version:
                    result_set.add(card_name)
    return sorted(list(result_set))


__all__ = (
    "register_base_class",
    "register_class",
    "get_instance",
    "get_class_list_by_base_class",
)
